# --------------------------------- table selection --------------------------------- #

BASELINE_TEXT2SQL_TABLE_SELECTION_PROMPT = """
### Instruction: Respond only with valid JSON. No introduction or summary needed.
Do not add ``` at start / end of the output. You will be given a database
schema and user query. Your job is to output the list of table names which will
be included in the user's SQL query. 

Here are some examples:
CREATE TABLE customers (
    customer_id INT,
    customer_name VARCHAR(100),
    contact_info VARCHAR(100)
);
CREATE TABLE orders (
    order_id INT,
    customer_id INT,
    order_date DATE,
    total_amount FLOAT
);
CREATE TABLE products (
    product_id INT,
    product_name VARCHAR(100),
    price FLOAT
);
User Query: "What are all the tables in database"
Output:
{{
    "include": ["customers", "orders", "products"]
}}

Example:
Schema:
CREATE TABLE employees (
    employee_id INT,
    employee_name VARCHAR(100),
    department VARCHAR(100),
    salary FLOAT
);
CREATE TABLE departments (
    department_id INT,
    department_name VARCHAR(100),
    location VARCHAR(100)
);
CREATE TABLE projects (
    project_id INT,
    project_name VARCHAR(100),
    budget FLOAT
);
User Query: "List the names of employees and their salaries."
Output:
{{
    "include": ["employees"]
}}
Example:
Schema:
CREATE TABLE students (
    student_id INT,
    student_name VARCHAR(100),
    grade_level INT
);

CREATE TABLE courses (
    course_id INT,
    course_name VARCHAR(100),
    teacher_id INT
);

CREATE TABLE enrollments (
    student_id INT,
    course_id INT,
    enrollment_date DATE
);
User Query: "Show the list of students enrolled in courses."
Output:
{{
    "include": ["students", "enrollments"]
}}

Example:
Schema:
CREATE TABLE authors (
    author_id INT,
    author_name VARCHAR(100)
);

CREATE TABLE books (
    book_id INT,
    book_title VARCHAR(100),
    author_id INT
);

CREATE TABLE publishers (
    publisher_id INT,
    publisher_name VARCHAR(100)
);

CREATE TABLE book_sales (
    sale_id INT,
    book_id INT,
    sale_amount FLOAT
);

User Query: "Find the total sales for each book."
Output:
{{
    "include": ["books", "book_sales"]
}}

------ Assistant's turn ------

Like above examples, here is your DB schema: {schema}
some additional info about the columns (optional): {additional_info}
and the user question: {question}

NOTE: The name of the tables should always match with the name of
the tables present the above schema

Respond only with valid JSON. Do not write an introduction or summary. output:
"""

# --------------------------------- text to sql --------------------------------- #

BASELINE_TEXT2SQL_WORKER_PROMPT_NO_FEWSHOT = """
# Instruction: 
- You will be given a question and a database schema.
- You need to write a SQL query to answer the question.
Do not add ``` at start / end of the query. It should be a single line query in 
a single line (string format).
- Make sure the column names are correct and exists in the table
- For column names which has a space with it, make sure you have put `` in that column name

# Database and Table Schema:
{schemas}
{additional_knowledge}
# Question: {question}
# SQL:
"""

BASELINE_TEXT2SQL_WORKER_PROMPT = """
# Instruction: 
- You will be given a question and a database schema.
- You need to write a SQL query to answer the question.
Do not add ``` at start / end of the query. It should be a single line query in 
a single line (string format).
- Make sure the column names are correct and exists in the table
- For column names which has a space with it, make sure you have put `` in that column name

# Database and Table Schema:
{schemas}

{additional_knowledge}

# Here are some Examples on how to generate SQL statements and use column names:
{few_shot_examples}

# Question: {question}

# SQL:
"""

# --------------------------------- error handling --------------------------------- #

BASELINE_TEXT2SQL_WORKER_ERROR_HANDLING_PROMPT = """
{existing_prompt}

# Generated SQL: {sql}

## Error Message

{error_msg}

Carefully review the original question and error message, then rewrite the SQL query to address the identified issues. 
Ensure your corrected query uses correct column names, 
follows proper SQL syntax, and accurately answers the original question 
without introducing new errors.

# SQL: 
"""

# --------------------------------- analysis base --------------------------------- #

BASELINE_ANALYSIS_WORKER_PROMPT = """
### Instruction: You will receive a user question and a table. Do not add ``` at start / end of the output. 
Analyze the table and provide your analysis following the below structure

# Analysis: (This should contain the answer from user's question)
# Reasoning: (This should contain the reasoning behind the your answer or analysis)

Example 1:

table: 
| Country | Population | GDP   |
|---------|------------|-------|
| A       | 5M         | 50B   |
| B       | 10M        | 100B  |
| C       | 2M         | 10B   |

User Question: Which country has the highest GDP per capita?
Output:

# Analysis: Country A has the highest GDP per capita.
# Reasoning: Country A's GDP per capita is 10,000, higher than B (10,000) and C (5,000).


Example 2:

table:
| Month    | Product | Sales | Returns | Profit Margin (%) |
|----------|---------|-------|---------|-------------------|
| January  | A       | 1000  | 50      | 20                |
| January  | B       | 1500  | 30      | 15                |
| February | A       | 1200  | 20      | 22                |
| February | B       | 1300  | 40      | 18                |
| March    | A       | 1100  | 25      | 21                |
| March    | B       | 1400  | 35      | 16                |

User Question: Which product had the highest profit margin in February?
Output:

# Analysis: Product A had the highest profit margin in February.
# Reasoning: Product A had a profit margin of 22% compared to Product B's 18% in February.


------ Assistant's turn ------

Dataframe to analyse: {dataframe}
user question: {question}
"""

# --------------------------------- analysis merger --------------------------------- #

BASELINE_ANALYSIS_MERGER_PROMPT = """Your task is to summarise
You will be given a set of summaries and reasoning in the form of a list of
json. Your task is to review the analuse and reasoning and summarise them:

- analysis1
- analysis2
- analysis3
- analysis4
and so on....

You will see the analysis and summarise them in a good
human readible format. 

------ Assistant's turn ------

Here is your analysis: {analysis}
Summary output:
"""

# --------------------------------- plot base --------------------------------- #

BASELINE_CHART_WORKER_PROMPT_TEMPLATE = """
### Instruction: Respond only with valid JSON. No introduction or summary needed. You are a senior data analyst. You know how to plot data when asked some analysis question. Do not add ``` at start / end of the output.

You will be given a user question, a list of dataframe column names, and you will output a JSON with
the following structure:

{{
    "x": # output the column name which should be on x-axis,
    "y": # output the column name which should be on y-axis,
    "plot_type": # The type of plot
}}

You can choose from these plot types:

- area: if you think you need to plot an area chart,
- bar: if you think you need to plot a bar chart,
- scatter: if you think you need to plot a scatter plot,
- histogram: if you think you need to plot a histogram,
- line: if you think you need to plot a line chart,
### Examples:

Example 1:
User Question: "Show the relationship between sales and revenue."

Dataframe columns: ["product_name", "sales", "revenue"]

Output:
{{
    "x": "sales",
    "y": "revenue",
    "plot_type": "scatter"
}}

Example 2:
User Question: "Show the distribution of product sales."

Dataframe columns: ["product_name", "sales"]

Output:
{{
    "x": "product_name",
    "y": "sales",
    "plot_type": "bar"
}}

Example 3:
User Question: "Show the change in temperature over time."

Dataframe columns: ["date", "temperature"]

Output:
{{
    "x": "date",
    "y": "temperature",
    "plot_type": "line"
}}

Example 4:
User Question: "Show the frequency distribution of employee ages."

Dataframe columns: ["employee_name", "age"]

Output:
{{
    "x": "age",
    "y": None,
    "plot_type": "histogram"
}}

------ Assistant's turn ------

Like the above example, here is your dataframe columns: {columns}
and here is user question: {question}

NOTE: From user's question you need to find the column of the table that makes
sense. Do not add any column name which is not present in {columns}. If there is nothing
to add just put None

Respond only with valid JSON. Do not write an introduction or summary. output:
"""

# --------------------------------- followup base --------------------------------- #

BASELINE_FOLLOWUP_WORKER_PROMPT = """
### Instruction: Respond only with valid JSON. No introduction or summary needed.
You will act like a simple conversation agent who knows to manage it's assistant who either plots or
write SQL query. If your assistant fails in quering the DB or plotting a graph then your job is to 
suggest user an alternative question or decision that will help your assistant to be successful in either query / plot / analysis next time.
Do not add ``` at start / end of the output.

You will be recieving the following input:
1. DB Schema: The schema of the database 
2. Decision: The decision that your assistant took. Decision could be either 'query' or 'plot' or 'analyse'
3. Query: The question that user asked.
4. Dataframe: (Optinal can be null) the dataframe that was being output or given as input
5. Analysis: Some text analysis from the model
6. Error from assistant: The error that your assistant made. 

Your job is to output a JSON with the following structure:

{{
    "alternate_decision": # can be either query / plot,
    "suggestion": # the alternate suggestive question 
}}

Here is an example:
Example 1:
DB Schema: 
CREATE TABLE sales (
    product_id INT,
    product_name VARCHAR(100),
    sale_date DATE,
    revenue FLOAT
);

Decision: "query"
Query: "What is the average revenue for products sold after 2023?"
Dataframe: null
Analysis: null
Error from assistant: "Query failed due to incorrect date format."

Output:
{{
    "alternate_decision": "query",
    "suggestion": "Could you find the average revenue for products sold in 2024 onwards?"
}}
Example 2:
DB Schema:
CREATE TABLE employees (
    employee_id INT,
    employee_name VARCHAR(100),
    department VARCHAR(100),
    salary FLOAT
);

Decision: "plot"
Query: "Show a line chart of the salary distribution by department."
Dataframe: null
Analysis: null
Error from assistant: "Data insufficient to generate the plot."

Output:
{{
    "alternate_decision": "plot",
    "suggestion": "Could you plot a bar chart showing total salary by department instead?"
}}

Example 1:
DB Schema:
CREATE TABLE sales (
    product_id INT,
    product_name VARCHAR(100),
    sale_date DATE,
    revenue FLOAT
);

Decision: "plot"
Query: "Plot the monthly revenue trend for 2024."
Dataframe: null
Analysis: null
Error from assistant: "Insufficient data to plot the monthly trend."

Output:
{{
    "alternate_decision": "query",
    "suggestion": "Can you query the total monthly revenue for 2024 so we can plot the trend afterward?"
}}

Example:
DB Schema:
CREATE TABLE sales (
    product_id INT,
    product_name VARCHAR(100),
    sale_date DATE,
    revenue FLOAT
);

Decision: "analyse"
Query: "Why xyz product name is doing unwell"
Dataframe: null
Analysis: "because they have less interaction, not good product"
Error from assistant: "null"

Output:
{{
    "alternate_decision": "analyse",
    "suggestion": "Why the product name xyz is not doing good and explain it in points"
}}

------ Assistant's turn ------

DB Schema: {schema}
Decision: {decision}
Query: {question}
Dataframe: {dataframe}
Analysis: {analysis}
Error from assistant: {error_from_model}

Your JSON keys should be only: `alternate_decision` and `suggestion`
Respond only with valid JSON. Do not write an introduction or summary. output:
"""



#!/usr/bin/env python
"""Django's command-line utility for administrative tasks."""
import os
import sys


def main():
    """Run administrative tasks."""
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "backend.settings")
    try:
        from django.core.management import execute_from_command_line
        import django
        # Patch Django's CommandParser before executing command
        from django.core.management.base import CommandParser
        original_init = CommandParser.__init__
        def new_init(self, **kwargs):
            kwargs.pop('allow_abbrev', None)  # Remove allow_abbrev if present
            original_init(self, **kwargs)
        CommandParser.__init__ = new_init
    except ImportError as exc:
        raise ImportError(
            "Couldn't import Django. Are you sure it's installed and "
            "available on your PYTHONPATH environment variable? Did you "
            "forget to activate a virtual environment?"
        ) from exc
    execute_from_command_line(sys.argv)


if __name__ == "__main__":
    main()



from django.contrib import admin
from django.urls import include, path
from drf_yasg import openapi
from drf_yasg.views import get_schema_view
from rest_framework import permissions

schema_view = get_schema_view(
    openapi.Info(
        title="PremSQL API",
        default_version="v0.0.1",
        description="API which controls PremSQL pipelines and agents",
        contact=openapi.Contact(email="anindyadeep@premai.io"),
        license=openapi.License(name="MIT"),
    ),
    public=True,
    permission_classes=(permissions.AllowAny,),
)

urlpatterns = [
    path("admin/", admin.site.urls),
    path("api/", include("api.urls")),
    path(
        "swagger<format>/", schema_view.without_ui(cache_timeout=0), name="schema-json"
    ),
    path(
        "swagger/",
        schema_view.with_ui("swagger", cache_timeout=0),
        name="schema-swagger-ui",
    ),
    path("redoc/", schema_view.with_ui("redoc", cache_timeout=0), name="schema-redoc"),
]



# Generated by Django 5.1.2 on 2024-10-26 09:06

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Session",
            fields=[
                ("session_id", models.AutoField(primary_key=True, serialize=False)),
                ("db_connection_uri", models.URLField()),
                ("session_name", models.CharField(max_length=255, unique=True)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("base_url", models.URLField()),
                ("session_db_path", models.CharField(max_length=255)),
            ],
            options={
                "ordering": ["created_at"],
            },
        ),
        migrations.CreateModel(
            name="Completions",
            fields=[
                ("chat_id", models.AutoField(primary_key=True, serialize=False)),
                ("message_id", models.IntegerField(blank=True, null=True)),
                ("session_name", models.CharField(max_length=255)),
                ("created_at", models.DateTimeField()),
                ("question", models.TextField(blank=True, null=True)),
                (
                    "session",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="messages",
                        to="api.session",
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "Completions",
                "ordering": ["-created_at"],
            },
        ),
    ]



import json
import os
import sqlite3
from abc import ABC, abstractmethod
from copy import deepcopy
from pathlib import Path
from typing import Optional, Sequence, Union

from tqdm.auto import tqdm

from premsql.logger import setup_console_logger
from premsql.prompts import BASE_TEXT2SQL_PROMPT
from premsql.utils import (
    filter_options,
    get_accepted_filters,
    get_random_few_shot_prompts,
    tokenize_fn,
)

logger = setup_console_logger(name="[DATASET]")

try:
    import torch
    from transformers import AutoTokenizer
except ImportError:
    logger.warn("Ensure transformers and torch. Install using: pip install torch transformers")


IGNORE_INDEX = -100


class Text2SQLBaseInstance:
    def __init__(self, dataset: list[dict]) -> None:

        assert "question" in dataset[0], "question is required"
        assert "SQL" in dataset[0], "sql is required"
        assert "db_path" in dataset[0], "db_path is required"
        assert "db_id" in dataset[0], "db_id is required"

        self.dataset = dataset

    def __repr__(self) -> str:
        return str(json.dumps(self.dataset[:3], indent=4))

    def __len__(self) -> int:
        return len(self.dataset)

    def __getitem__(self, idx: int) -> dict:
        return dict(**self.dataset[idx])

    def schema_prompt(self, db_path: str) -> str:
        schemas = {}
        conn = sqlite3.connect(db_path)
        cursor = conn.cursor()
        cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
        tables = cursor.fetchall()

        for table in tables:
            table_name = table[0]
            if table_name == "sqlite_sequence":
                continue
            cursor.execute(
                f"SELECT sql FROM sqlite_master WHERE type='table' AND name='{table_name}';"
            )
            create_table_sql = cursor.fetchone()
            if create_table_sql:
                schemas[table_name] = create_table_sql[0]
            else:
                schemas[table_name] = "Schema does not exist"

        schema_prompt = "\n".join(
            schemas[table[0]] for table in tables if table[0] != "sqlite_sequence"
        )
        return schema_prompt

    def additional_prompt(self, prompt: Optional[str] = None):
        return "" if prompt is None else f"# Additional Knowledge:\n{prompt}"

    def add_few_shot_examples(self, db_id: str, k: int = 3) -> str:
        assert k > 0, "k should be greater than 0"
        db_fewshot_prompts_map = get_random_few_shot_prompts(
            dataset=self.dataset, num_few_shot=k
        )
        return db_fewshot_prompts_map[db_id]

    def apply_prompt(
        self,
        num_fewshot: Optional[int] = None,
        prompt_template: Optional[str] = None,
    ):
        prompt_template = (
            BASE_TEXT2SQL_PROMPT if prompt_template is None else prompt_template
        )
        for blob in tqdm(self.dataset, total=len(self.dataset), desc="Applying prompt"):
            few_shot_prompt = (
                ""
                if num_fewshot is None
                else self.add_few_shot_examples(db_id=blob["db_id"], k=num_fewshot)
            )
            final_prompt = prompt_template.format(
                schemas=self.schema_prompt(blob["db_path"]),
                additional_knowledge=(
                    ""
                    if "knowledge" not in blob
                    else self.additional_prompt(blob["knowledge"])
                ),
                few_shot_examples=few_shot_prompt,
                question=blob["question"],
            )
            blob["prompt"] = final_prompt
        return self.dataset


class SupervisedDatasetForTraining(torch.utils.data.Dataset):
    @classmethod
    def load_from_pth(cls, dataset_path: Union[str, Path]):
        dataset_path = str(dataset_path)
        dataset_dict = torch.load(dataset_path)

        assert "input_ids" in dataset_dict[0], "input_ids is required"
        assert "labels" in dataset_dict[0], "labels is required"
        assert "raw" in dataset_dict[0], "raw is required"

        return cls(
            dataset=dataset_dict,
            model_name_or_path=None,
            hf_token=None,
        )

    def __init__(
        self,
        dataset: dict,
        model_name_or_path: Optional[str] = None,
        tokenize: Optional[bool] = False, 
        hf_token: Optional[str] = None,
    ):
        assert "prompt" in dataset[0], "key prompt is required"
        assert "SQL" in dataset[0], "key SQL is required"

        self.is_tokenized = False

        if model_name_or_path is not None and tokenize:
            self.tokenizer = AutoTokenizer.from_pretrained(
                pretrained_model_name_or_path=model_name_or_path,
                padding_side="right",
                token=hf_token,
            )
            self.dataset = dataset

            if self.tokenizer.chat_template:
                for content in self.dataset:
                    content["prompt"] = self.tokenizer.apply_chat_template(
                        [{"role": "user", "content": content["prompt"]}], tokenize=False
                    )
                logger.info("Casted dataset with model chat template")

            logger.info("Starting Tokenization ...")
            sources, targets = [], []
            for example in self.dataset:
                sources.append(example["prompt"])
                targets.append(f"{example['SQL']}{self.tokenizer.eos_token}")

            data_dict = self.preprocess(sources=sources, targets=targets)
            self.input_ids = data_dict["input_ids"]
            self.labels = data_dict["labels"]
            self.is_tokenized = True

        elif "input_ids" in dataset[0] and "labels" in dataset[0]:
            self.dataset = dataset
            self.input_ids = dataset["input_ids"]
            self.labels = dataset["labels"]
            self.is_tokenized = True

        elif model_name_or_path is not None and not tokenize:
            self.tokenizer = AutoTokenizer.from_pretrained(
                pretrained_model_name_or_path=model_name_or_path,
                padding_side="right",
                token=hf_token,
            )
            self.dataset = dataset
            if self.tokenizer.chat_template:
                for content in self.dataset:
                    content["prompt"] = self.tokenizer.apply_chat_template(
                        [{"role": "user", "content": content["prompt"]}], tokenize=False
                    )
                logger.info("Casted dataset with model chat template")
        else:
            self.dataset = dataset

    def preprocess(self, sources: Sequence[str], targets: Sequence[str]):
        examples = [s + t for s, t in zip(sources, targets)]
        examples_tokenized, sources_tokenized = [
            tokenize_fn(strings, self.tokenizer) for strings in (examples, sources)
        ]
        input_ids = examples_tokenized["input_ids"]
        labels = deepcopy(input_ids)

        for label, source_len in zip(labels, sources_tokenized["input_ids_lens"]):
            label[:source_len] = IGNORE_INDEX

        return dict(input_ids=input_ids, labels=labels)

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx: int):
        if self.is_tokenized:
            return dict(
                input_ids=self.input_ids[idx],
                labels=self.labels[idx],
                raw=dict(**self.dataset[idx]),
            )
        else:
            return dict(**self.dataset[idx])

    def save_tokenized_dataset(self, path_to_save: Union[str, Path]):
        torch.save(self.dataset, str(path_to_save))
        logger.info("Dataset saved successfully in {}".format(path_to_save))


class Text2SQLBaseDataset(ABC):
    def __init__(
        self,
        split: str,
        dataset_path: Union[str, Path],
        database_folder_name: str,
        json_file_name: str,
        hf_token: Optional[str] = None,
    ):
        self.dataset_path = Path(dataset_path)
        self.database_folder_name = database_folder_name
        self.dataset = json.load(open(self.dataset_path / json_file_name, "r"))
        assert split in ["train", "validation", "test"], ValueError(
            "Split should be either train or validation"
        )
        self.split = split
        self.hf_token = hf_token if hf_token else os.environ.get("HF_TOKEN", None)

    @property
    def raw_dataset(self):
        return self._text2sql_dataset.dataset

    @property
    def filter_availables(self):
        return get_accepted_filters(data=self.dataset)

    @abstractmethod
    def setup_dataset(
        self,
        filter_by: Optional[tuple] = None,
        num_rows: Optional[int] = None,
        num_fewshot: Optional[int] = None,
        model_name_or_path: Optional[str] = None,
        tokenize: Optional[bool] = False,
        prompt_template: Optional[str] = BASE_TEXT2SQL_PROMPT,
    ):
        for content in self.dataset:
            content["db_path"] = str(
                self.dataset_path
                / f"{self.database_folder_name}"
                / content["db_id"]
                / f"{content['db_id']}.sqlite"
            )

        if filter_by:
            self.dataset = filter_options(data=self.dataset, filter_by=filter_by)

        if num_rows:
            self.dataset = self.dataset[:num_rows]

        self.dataset = Text2SQLBaseInstance(dataset=self.dataset).apply_prompt(
            num_fewshot=num_fewshot, prompt_template=prompt_template
        )
        return SupervisedDatasetForTraining(
            dataset=self.dataset,
            model_name_or_path=model_name_or_path,
            hf_token=self.hf_token,
            tokenize=tokenize
        )

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx):
        return dict(**self.dataset[idx])


class StandardDataset(Text2SQLBaseDataset):
    def __init__(
        self,
        split: str,
        dataset_path: Union[str, Path],
        database_folder_name: str,
        json_file_name: str,
        hf_token: Optional[str] = None,
    ):
        super().__init__(
            split=split,
            dataset_path=dataset_path,
            database_folder_name=database_folder_name,
            json_file_name=json_file_name,
            hf_token=hf_token,
        )

    def setup_dataset(
        self,
        filter_by: tuple | None = None,
        num_rows: int | None = None,
        num_fewshot: int | None = None,
        model_name_or_path: str | None = None,
        prompt_template: str | None = None,
        tokenize: bool | None = False 
    ):
        logger.info("Setting up Dataset")
        return super().setup_dataset(
            filter_by=filter_by,
            num_rows=num_rows,
            model_name_or_path=model_name_or_path,
            tokenize=tokenize,
            prompt_template=prompt_template,
            num_fewshot=num_fewshot
        )



from typing import Any, Dict, Literal

import pandas as pd

from premsql.executors.from_langchain import SQLDatabase
from premsql.logger import setup_console_logger
from premsql.agents.models import AgentOutput, ExitWorkerOutput

logger = setup_console_logger("[PIPELINE-UTILS]")


def convert_df_to_dict(df: pd.DataFrame):
    return {"data": df.to_dict(), "columns": list(df.keys())}


def execute_and_render_result(
    db: SQLDatabase, sql: str, using: Literal["dataframe", "json"]
):
    result = db.run_no_throw(command=sql, fetch="cursor")

    if isinstance(result, str):
        return _render_error(result, sql, using)
    return _render_data(result, sql, using)


def _render_error(error: str, sql: str, using: str) -> Dict[str, Any]:
    to_show = {"sql_string": sql, "error_from_model": error, "dataframe": None}

    if using == "dataframe":
        to_show["dataframe"] = pd.DataFrame()  # empty DataFrame
    elif using == "json":
        to_show["dataframe"] = {"data": {}, "columns": []}  # empty JSON structure
    return to_show


def _render_data(result, sql: str, using: str) -> Dict[str, Any]:
    table = pd.DataFrame(data=result.fetchall(), columns=result.keys())
    if len(table) > 200:
        logger.info("Truncating output table to first 200 rows only")
        table = table.iloc[:200, :]
    
    if any(table.columns.duplicated()):
        logger.info(f"Found duplicate columns: {table.columns[table.columns.duplicated()].tolist()}")
        # Create unique column names by adding suffixes
        table.columns = [f"{col}_{i}" if i > 0 else col 
                        for i, col in enumerate(table.columns)]
        logger.info(f"Renamed columns to: {table.columns.tolist()}")

    to_show = {"sql_string": sql, "error_from_model": None, "dataframe": table}

    if using == "json":
        to_show["dataframe"] = {"columns": list(table.columns), "data": table.to_dict()}
    return to_show



def convert_exit_output_to_agent_output(exit_output: ExitWorkerOutput) -> AgentOutput:
    return AgentOutput(
        session_name=exit_output.session_name,
        question=exit_output.question,
        db_connection_uri=exit_output.db_connection_uri,
        route_taken=exit_output.route_taken,
        input_dataframe=exit_output.sql_input_dataframe
        or exit_output.analysis_input_dataframe
        or exit_output.plot_input_dataframe,
        output_dataframe=exit_output.sql_output_dataframe
        or exit_output.plot_output_dataframe,
        sql_string=exit_output.sql_string,
        analysis=exit_output.analysis,
        reasoning=exit_output.sql_reasoning
        or exit_output.analysis_reasoning
        or exit_output.plot_reasoning,
        plot_config=exit_output.plot_config,
        image_to_plot=exit_output.image_to_plot,
        followup_route=exit_output.followup_route_to_take,
        followup_suggestion=exit_output.followup_suggestion,
        error_from_pipeline=(
            exit_output.error_from_sql_worker
            or exit_output.error_from_analysis_worker
            or exit_output.error_from_plot_worker
            or exit_output.error_from_followup_worker
        ),
    )


from typing import Optional

import pandas as pd

from premsql.generators.base import Text2SQLGeneratorBase
from premsql.logger import setup_console_logger
from premsql.agents.base import ChartPlotWorkerBase, ChartPlotWorkerOutput
from premsql.agents.baseline.prompts import BASELINE_CHART_WORKER_PROMPT_TEMPLATE
from premsql.agents.tools.plot.base import BasePlotTool
from premsql.agents.utils import convert_df_to_dict

logger = setup_console_logger("[PLOT-WORKER]")


class BaseLinePlotWorker(ChartPlotWorkerBase):
    def __init__(
        self, generator: Text2SQLGeneratorBase, plot_tool: BasePlotTool
    ) -> None:
        self.generator, self.plot_tool = generator, plot_tool

    def run(
        self,
        question: str,
        input_dataframe: pd.DataFrame,
        temperature: Optional[float] = 0.1,
        max_new_tokens: Optional[int] = 100,
        plot_image: Optional[bool] = True,
        prompt_template: Optional[str] = BASELINE_CHART_WORKER_PROMPT_TEMPLATE,
        **kwargs,
    ) -> ChartPlotWorkerOutput:
        prompt = prompt_template.format(
            columns=list(input_dataframe.columns), question=question
        )
        try:
            logger.info("Going for generation")
            to_plot = self.generator.generate(
                data_blob={"prompt": prompt},
                temperature=temperature,
                max_new_tokens=max_new_tokens,
                postprocess=False,
            )
            to_plot = to_plot.replace("null", "None")
            plot_config = eval(to_plot)
            fig = self.plot_tool.run(data=input_dataframe, plot_config=plot_config)
            logger.info(f"Plot config: {plot_config}")

            if plot_image:
                output = self.plot_tool.convert_image_to_base64(
                    self.plot_tool.convert_plot_to_image(fig=fig)
                )
                logger.info("Done base64 conversion")
            else:
                output = None

            return ChartPlotWorkerOutput(
                question=question,
                input_dataframe=convert_df_to_dict(input_dataframe),
                plot_config=plot_config,
                plot_reasoning=None,
                output_dataframe=None,
                image_plot=output,
                error_from_model=None,
                additional_input={
                    "temperature": temperature,
                    "max_new_tokens": max_new_tokens,
                    **kwargs,
                },
            )

        except Exception as e:
            error_message = f"Error during plot generation: {str(e)}"
            return ChartPlotWorkerOutput(
                question=question,
                input_dataframe=convert_df_to_dict(input_dataframe),
                plot_config=None,
                image_plot=None,
                plot_reasoning=None,
                error_from_model=error_message,
                additional_input={
                    "temperature": temperature,
                    "max_new_tokens": max_new_tokens,
                    **kwargs,
                },
            )



import time
from typing import Union

from langchain_community.utilities.sql_database import SQLDatabase

from premsql.executors.base import BaseExecutor
from premsql.utils import convert_sqlite_path_to_dsn


class ExecutorUsingLangChain(BaseExecutor):

    def execute_sql(self, sql: str, dsn_or_db_path: Union[str, SQLDatabase]) -> dict:
        if isinstance(dsn_or_db_path, str):
            if dsn_or_db_path.endswith("sqlite"):
                dsn_or_db_path = convert_sqlite_path_to_dsn(path=dsn_or_db_path)
            db = SQLDatabase.from_uri(dsn_or_db_path)
        else:
            db = dsn_or_db_path

        start_time = time.time()
        response = db.run_no_throw(sql)
        end_time = time.time()

        error = response if response.startswith("Error") else None
        return {
            "result": None if error else response,
            "error": error,
            "execution_time": end_time - start_time,
        }



import sqlite3
import time

from contextlib import contextmanager
from typing import Any, Dict, Generator 
from premsql.executors.base import BaseExecutor
from premsql.logger import setup_console_logger


class OptimizedSQLiteExecutor(BaseExecutor):
    def __init__(self, timeout: float = 1000.0) -> None:
        self.timeout = timeout
        self.logger = setup_console_logger(name="[OPTIMIZED-SQLite-EXEC]")

    @contextmanager
    def get_connection(self, db_path: str) -> Generator[sqlite3.Connection, None, None]:
        if db_path.startswith("sqlite:///"):
            db_path = db_path.split("sqlite:///")[1]
        
        conn = sqlite3.connect(db_path, timeout=self.timeout)
        conn.execute("PRAGMA journal_mode = WAL")
        conn.execute("PRAGMA synchronous = NORMAL")
        conn.execute("PRAGMA cache_size = -64000")  # 64MB cache
        conn.execute("PRAGMA temp_store = MEMORY")
        conn.row_factory = sqlite3.Row
        try:
            yield conn 
        finally:
            conn.close() 
        
    def execute_sql(self, sql: str, dsn_or_db_path: str) -> Dict[str, Any]:
        start_time = time.time()
        try:
            with self.get_connection(dsn_or_db_path) as conn:
                cursor = conn.cursor()
                cursor.execute("EXPLAIN QUERY PLAN " + sql)
                query_plan = cursor.fetchall()
                
                if any("SCAN TABLE" in str(row) for row in query_plan):
                    self.logger.warn("Warning: Full table scan detected. Consider adding an index.")
                
                cursor.execute(sql)
                result = [dict(row) for row in cursor.fetchall()]
                error = None
        except sqlite3.Error as e:
            result = None
            error = str(e)
        finally:
            end_time = time.time()

        return {
            "result": result,
            "error": error,
            "execution_time": end_time - start_time,
        }
    
    def match_sqls(self, predicted_sql: str, gold_sql: str, dsn_or_db_path: str) -> Dict[str, Any]:
        with self.get_connection(dsn_or_db_path) as conn:
            prediction = self.execute_sql(predicted_sql, dsn_or_db_path)
            gold = self.execute_sql(gold_sql, dsn_or_db_path)

        if prediction["error"]:
            return {"result": 0, "error": prediction["error"]}

        is_match = set(map(tuple, prediction["result"])) == set(map(tuple, gold["result"]))
        return {
            "result": int(is_match),
            "error": None if is_match else "Table mismatch",
        }

    def iterated_execution(self, predicted_sql: str, gold_sql: str, dsn_or_db_path: str, num_iterations: int) -> Dict[str, Any]:
        is_match = self.match_sqls(predicted_sql, gold_sql, dsn_or_db_path)

        if is_match["result"] == 1:
            with self.get_connection(dsn_or_db_path) as conn:
                diff_list = []
                for _ in range(num_iterations):
                    gold_time = self.execute_sql(gold_sql, dsn_or_db_path)["execution_time"]
                    predicted_time = self.execute_sql(predicted_sql, dsn_or_db_path)["execution_time"]
                    diff_list.append(predicted_time / gold_time if gold_time > 0 else float('inf'))

            processed_diff_list = self.clean_abnormal(diff_list)
            return {
                "result": sum(processed_diff_list) / len(processed_diff_list) if processed_diff_list else 0,
                "error": None,
            }
        return {"result": 0, "error": is_match["error"]}



class SQLiteExecutor(BaseExecutor):
    def execute_sql(self, sql: str, dsn_or_db_path: str) -> dict:
        if dsn_or_db_path.startswith("sqlite:///"):
            dsn_or_db_path = dsn_or_db_path.split("sqlite:///")[1]
            
        conn = sqlite3.connect(dsn_or_db_path)
        cursor = conn.cursor()

        start_time = time.time()
        try:
            cursor.execute(sql)
            result = cursor.fetchall()
            error = None
        except Exception as e:
            result = None
            error = str(e)

        end_time = time.time()
        cursor.close()
        conn.close()

        result = {
            "result": result,
            "error": error,
            "execution_time": end_time - start_time,
        }
        return result


import requests
from premsql.logger import setup_console_logger
from premsql.playground.backend.api.pydantic_models import (
    SessionCreationResponse,
    SessionDeleteResponse,
    SessionListResponse,
    SessionCreationRequest,
    CompletionCreationRequest,
    CompletionCreationResponse,
    CompletionListResponse,
)

BASE_URL = "http://127.0.0.1:8000/api"

logger = setup_console_logger("BACKEND-API-CLIENT")


class BackendAPIClient:
    def __init__(self):
        self.base_url = BASE_URL
        self.headers = {
            'accept': 'application/json',
            'Content-Type': 'application/json',
        }

    def create_session(self, request: SessionCreationRequest) -> SessionCreationResponse:
        try:
            response = requests.post(
                f"{self.base_url}/session/create",
                json=request.model_dump(),
                headers=self.headers
            )
            response.raise_for_status()  # Raises an HTTPError for bad responses
            
            return SessionCreationResponse(**response.json())
        except requests.RequestException as e:
            logger.error(f"Error creating session: {str(e)}")
            logger.error(f"Response content: {response.text if 'response' in locals() else 'No response'}")
            return SessionCreationResponse(
                status="error",
                status_code=response.status_code if 'response' in locals() and hasattr(response, 'status_code') else 500,
                error_message=f"Failed to create session: {str(e)}"
            )
        except ValueError as e:
            logger.error(f"Error parsing response: {str(e)}")
            logger.error(f"Response content: {response.text if 'response' in locals() else 'No response'}")
            return SessionCreationResponse(
                status="error",
                status_code=500,
                error_message=f"Failed to parse server response: {str(e)}"
            )

        except requests.RequestException as e:
            logger.error(f"Error creating session: {str(e)}")
            logger.error(f"Response content: {response.text}")
            return SessionCreationResponse(
                status="error",
                status_code=response.status_code if hasattr(response, 'status_code') else 500,
                error_message=f"Failed to create session: {str(e)}"
            )
        except ValueError as e:
            logger.error(f"Error parsing response: {str(e)}")
            logger.error(f"Response content: {response.text}")
            return SessionCreationResponse(
                status="error",
                status_code=500,
                error_message=f"Failed to parse server response: {str(e)}"
            )

    def list_sessions(self, page: int = 1, page_size: int = 20) -> SessionListResponse:
        try:
            response = requests.get(
                f"{self.base_url}/session/list/",
                params={"page": page, "page_size": page_size},
                headers=self.headers
            )
            response.raise_for_status()
            return SessionListResponse(**response.json())
        except requests.RequestException as e:
            logger.error(f"Error listing sessions: {str(e)}")
            logger.error(f"Response content: {response.text if 'response' in locals() else 'No response'}")
            return SessionListResponse(
                status="error",
                status_code=response.status_code if 'response' in locals() and hasattr(response, 'status_code') else 500,
                error_message=f"Failed to list sessions: {str(e)}",
                sessions=[],
                total_count=0
            )
        except ValueError as e:
            logger.error(f"Error parsing response: {str(e)}")
            logger.error(f"Response content: {response.text if 'response' in locals() else 'No response'}")
            return SessionListResponse(
                status="error",
                status_code=500,
                error_message=f"Failed to parse server response: {str(e)}",
                sessions=[],
                total_count=0
            )

    def get_session(self, session_name: str) -> SessionListResponse:
        try:
            response = requests.get(
                f"{self.base_url}/session/{session_name}/",
                headers=self.headers
            )
            response.raise_for_status()
            return SessionListResponse(**response.json())
        except requests.RequestException as e:
            logger.error(f"Error getting session: {str(e)}")
            logger.error(f"Response content: {response.text if 'response' in locals() else 'No response'}")
            return SessionListResponse(
                status="error",
                status_code=response.status_code if 'response' in locals() and hasattr(response, 'status_code') else 500,
                error_message=f"Failed to get session: {str(e)}",
                name="",
                created_at="",
                sessions=[]
            )
        except (ValueError, KeyError, IndexError) as e:
            logger.error(f"Error parsing response: {str(e)}")
            logger.error(f"Response content: {response.text if 'response' in locals() else 'No response'}")
            return SessionListResponse(
                status="error",
                status_code=500,
                error_message=f"Failed to parse server response: {str(e)}",
                name="",
                created_at="",
                sessions=[]
            )

    def delete_session(self, session_name: str) -> SessionDeleteResponse:
        try:
            response = requests.delete(
                f"{self.base_url}/session/{session_name}",
                headers=self.headers
            )
            response.raise_for_status()
            return SessionDeleteResponse(**response.json())
        except requests.RequestException as e:
            logger.error(f"Error deleting session: {str(e)}")
            logger.error(f"Response content: {response.text if 'response' in locals() else 'No response'}")
            return SessionDeleteResponse(
                status="error",
                status_code=response.status_code if 'response' in locals() and hasattr(response, 'status_code') else 500,
                error_message=f"Failed to delete session: {str(e)}"
            )
        except ValueError as e:
            logger.error(f"Error parsing response: {str(e)}")
            logger.error(f"Response content: {response.text if 'response' in locals() else 'No response'}")
            return SessionDeleteResponse(
                status="error",
                status_code=500,
                error_message=f"Failed to parse server response: {str(e)}"
            )

    # Chats
    def create_completion(self, request: CompletionCreationRequest) -> CompletionCreationResponse:
        try:
            response = requests.post(
                f"{self.base_url}/chat/completion",
                json=request.model_dump(),
                headers=self.headers
            )
            response.raise_for_status()
            return CompletionCreationResponse(**response.json())
        except requests.RequestException as e:
            logger.error(f"Error creating completion: {str(e)}")
            logger.error(f"Response content: {response.text if 'response' in locals() else 'No response'}")
            return CompletionCreationResponse(
                status="error",
                status_code=response.status_code if 'response' in locals() and hasattr(response, 'status_code') else 500,
                error_message=f"Failed to create completion: {str(e)}",
                completion=""
            )
        except ValueError as e:
            logger.error(f"Error parsing response: {str(e)}")
            logger.error(f"Response content: {response.text if 'response' in locals() else 'No response'}")
            return CompletionCreationResponse(
                status="error",
                status_code=500,
                error_message=f"Failed to parse server response: {str(e)}",
                completion=""
            )
    
    def get_chat_history(self, session_name: str, page: int = 1, page_size: int = 20) -> CompletionListResponse:
        try:
            response = requests.get(
                f"{self.base_url}/chat/history/{session_name}/",
                params={"page": page, "page_size": page_size},
                headers=self.headers
            )
            response.raise_for_status()
            return CompletionListResponse(**response.json())
        except requests.RequestException as e:
            logger.error(f"Error getting chat history: {str(e)}")
            logger.error(f"Response content: {response.text if 'response' in locals() else 'No response'}")
            return CompletionListResponse(
                status="error",
                status_code=500,
                error_message=f"Failed to get chat history: {str(e)}",
                completions=[],
                total_count=0
            )
        except ValueError as e:
            logger.error(f"Error parsing response: {str(e)}")
            logger.error(f"Response content: {response.text if 'response' in locals() else 'No response'}")
            return CompletionListResponse(
                status="error",
                status_code=500,
                error_message=f"Failed to parse server response: {str(e)}",
                completions=[],
                total_count=0
            )



BASE_TEXT2SQL_PROMPT = """
# Follow these instruction:
You will be given schemas of tables of a database. Your job is to write correct
error free SQL query based on the question asked. Please make sure:

1. Do not add ``` at start / end of the query. It should be a single line query in a  single line (string format)
2. Make sure the column names are correct and exists in the table
3. For column names which has a space with it, make sure you have put `` in that column name
4. Think step by step and always check schema and question and the column names before writing the
query. 

# Database and Table Schema:
{schemas}

{additional_knowledge}

# Here are some Examples on how to generate SQL statements and use column names:
{few_shot_examples}

# Question: {question}

# SQL: 
"""

OLD_BASE_TEXT2SQL_PROMPT = """
# Instruction: 
- You will be given a question and a database schema.
- You need to write a SQL query to answer the question.
Do not add ``` at start / end of the query. It should be a single line query in 
a single line (string format).
- Make sure the column names are correct and exists in the table
- For column names which has a space with it, make sure you have put `` in that column name

# Database and Table Schema:
{schemas}

{additional_knowledge}

# Here are some Examples on how to generate SQL statements and use column names:
{few_shot_examples}

# Question: {question}

# SQL:
"""

ERROR_HANDLING_PROMPT = """
{existing_prompt}

# Generated SQL: {sql}

## Error Message

{error_msg}

Carefully review the original question and error message, then rewrite the SQL query to address the identified issues. 
Ensure your corrected query uses correct column names, 
follows proper SQL syntax, and accurately answers the original question 
without introducing new errors.

# SQL: 
"""



import math
import traceback
from pathlib import Path
from typing import Optional, Union

from func_timeout import FunctionTimedOut, func_timeout
from tqdm.auto import tqdm

from premsql.executors.base import BaseExecutor
from premsql.utils import save_to_json


class Text2SQLEvaluator:
    def __init__(
        self, executor: BaseExecutor, experiment_path: Union[str, Path]
    ) -> None:
        self.executor = executor
        self.experiment_path = Path(experiment_path)

    def _execute_model(
        self,
        metric_name: str,
        generated_sql: str,
        gold_sql: str,
        dsn_or_db_path: str,
        meta_time_out: Optional[int] = 1000,
        num_iterations: Optional[int] = None,
        debug: Optional[bool] = False,
    ):
        assert metric_name in ["accuracy", "ves"], "Invalid metric name"
        try:
            if metric_name == "accuracy":
                result = func_timeout(
                    meta_time_out,
                    self.executor.match_sqls,
                    args=(generated_sql, gold_sql, dsn_or_db_path),
                )
            elif metric_name == "ves":
                num_iterations = 10 if num_iterations is None else num_iterations
                result = func_timeout(
                    meta_time_out,
                    self.executor.iterated_execution,
                    args=(generated_sql, gold_sql, dsn_or_db_path, num_iterations),
                )
            else:
                raise ValueError(f"Invalid metric name: {metric_name}")

            return {
                metric_name: result["result"],
                "error": result["error"],
            }
        except FunctionTimedOut as e:
            return {
                metric_name: 0,
                "error": f"Function Timed out: {e}",
            }
        except Exception as e:
            if debug:
                traceback.print_exc()

            return {
                metric_name: 0,
                "error": f"Exception: {e}",
            }

    def execute(
        self,
        metric_name: str,
        model_responses: list[dict],
        filter_by: Optional[str] = None,
        num_iterations: Optional[int] = 10,
        meta_time_out: Optional[int] = 10,  # change it later to 1000
        debug: Optional[bool] = False,
    ) -> dict:
        data_with_results = []

        for response in tqdm(model_responses, total=len(model_responses)):
            result = self._execute_model(
                metric_name=metric_name,
                generated_sql=response["generated"],
                gold_sql=response["SQL"],
                dsn_or_db_path=response["db_path"],
                num_iterations=num_iterations,
                meta_time_out=meta_time_out,
                debug=debug,
            )
            data_with_results.append({**response, **result})

        execution_result = {}
        if filter_by:
            if filter_by not in data_with_results[0]:
                raise KeyError(f"Filter key: {filter_by} is not found in responses")

            filter_values = {response[filter_by] for response in data_with_results}
            total_responses = len(data_with_results)
            overall_metric = 0.0

            for value in filter_values:
                filtered_responses = [
                    response
                    for response in data_with_results
                    if response[filter_by] == value
                ]
                metric_value = self.compute_metric(
                    results=filtered_responses, metric_name=metric_name
                )
                execution_result[value] = metric_value
                overall_metric += (
                    metric_value * len(filtered_responses) / total_responses
                )

            execution_result["overall"] = overall_metric
        else:
            execution_result["overall"] = self.compute_metric(
                results=data_with_results, metric_name=metric_name
            )

        save_to_json(
            json_object=execution_result,
            save_path=self.experiment_path / f"{metric_name}.json",
        )

        # also save the data_with_results
        save_to_json(
            json_object=data_with_results,
            save_path=self.experiment_path / "predict.json",
        )
        return execution_result

    def compute_metric(self, results: list[dict], metric_name: str) -> float:
        if metric_name == "accuracy":
            return sum(res["accuracy"] for res in results) / len(results) * 100

        elif metric_name == "ves":
            num_queries = len(results)
            total_ratio = 0.0
            for result in results:
                total_ratio += math.sqrt(result["ves"]) * 100
            ves = total_ratio / num_queries
            return ves

        else:
            raise ValueError(f"Invalid metric name: {metric_name}")



from abc import ABC, abstractmethod
from typing import Optional, Union

import pandas as pd

from premsql.executors.base import BaseExecutor
from premsql.executors.from_langchain import SQLDatabase
from premsql.generators.base import Text2SQLGeneratorBase
from premsql.logger import setup_console_logger
from premsql.agents.memory import AgentInteractionMemory
from premsql.agents.models import (
    AgentOutput,
    AnalyserWorkerOutput,
    ChartPlotWorkerOutput,
    ExitWorkerOutput,
    RouterWorkerOutput,
    Text2SQLWorkerOutput,
)

logger = setup_console_logger("[PIPELINE-BASE]")


# If someone wants to make a new worker class
class WorkerBase(ABC):
    @abstractmethod
    def run(self):
        return NotImplementedError()


class AnalysisWorkerBase(ABC):
    @abstractmethod
    def run(
        self, question: str, input_dataframe: Optional[pd.DataFrame] = None
    ) -> AnalyserWorkerOutput:
        raise NotImplementedError


class ChartPlotWorkerBase(ABC):
    @abstractmethod
    def run(
        self, question: str, input_dataframe: Optional[pd.DataFrame] = None
    ) -> ChartPlotWorkerOutput:
        raise NotImplementedError


class RouterWorkerBase(ABC):
    @abstractmethod
    def run(
        self, question: str, input_dataframe: Optional[pd.DataFrame] = None
    ) -> RouterWorkerOutput:
        raise NotImplementedError


class Text2SQLWorkerBase(ABC):
    def __init__(
        self,
        db_connection_uri: str,
        generator: Text2SQLGeneratorBase,
        executor: BaseExecutor,
        include_tables: Optional[str] = None,
        exclude_tables: Optional[str] = None,
    ) -> None:

        self.generator, self.executor = generator, executor
        self.db_connection_uri = db_connection_uri
        self.db = self.initialize_database(
            db_connection_uri=db_connection_uri,
            include_tables=include_tables,
            exclude_tables=exclude_tables,
        )

    @abstractmethod
    def run(self, question: str, **kwargs) -> Text2SQLWorkerOutput:
        raise NotImplementedError

    def initialize_database(
        self,
        db_connection_uri: str,
        include_tables: Optional[list] = None,
        exclude_tables: Optional[list] = None,
    ) -> SQLDatabase:
        """This method should return a db object

        To customise this method you make a different db object but
        it should have similar methods and behaviour like
        langchain SQLDatbase. You can find the implementation of SQLDatabase
        here: https://api.python.langchain.com/en/latest/_modules/langchain_community/utilities/sql_database.html#SQLDatabase
        """
        try:
            return SQLDatabase.from_uri(
                database_uri=db_connection_uri,
                sample_rows_in_table_info=0,
                ignore_tables=exclude_tables,
                include_tables=include_tables
            )
        except Exception as e:
            logger.error(f"Error loading the database: {e}")
            raise RuntimeError(f"Error loading the database: {e}")


class AgentBase(ABC):
    def __init__(
        self,
        session_name: str,
        db_connection_uri: str,
        session_db_path: Optional[str] = None,
        route_worker_kwargs: Optional[dict] = None,
    ) -> None:
        self.session_name, self.db_connection_uri = session_name, db_connection_uri
        self.history = AgentInteractionMemory(
            session_name=session_name, db_path=session_db_path
        )
        self.session_db_path = self.history.db_path
        self.route_worker_kwargs = route_worker_kwargs

    @abstractmethod
    def run(
        self,
        question: str,
        input_dataframe: Optional[dict] = None,
        server_mode: Optional[bool] = False,
    ) -> Union[ExitWorkerOutput, AgentOutput]:
        # Make sure you convert the dataframe to a table
        raise NotImplementedError()

    def convert_exit_output_to_agent_output(
        self, exit_output: ExitWorkerOutput
    ) -> AgentOutput:
        return AgentOutput(
            session_name=exit_output.session_name,
            question=exit_output.question,
            db_connection_uri=exit_output.db_connection_uri,
            route_taken=exit_output.route_taken,
            input_dataframe=exit_output.sql_input_dataframe
            or exit_output.analysis_input_dataframe
            or exit_output.plot_input_dataframe,
            output_dataframe=exit_output.sql_output_dataframe
            or exit_output.plot_output_dataframe,
            sql_string=exit_output.sql_string,
            analysis=exit_output.analysis,
            reasoning=exit_output.sql_reasoning
            or exit_output.analysis_reasoning
            or exit_output.plot_reasoning,
            plot_config=exit_output.plot_config,
            image_to_plot=exit_output.image_to_plot,
            followup_route=exit_output.followup_route_to_take,
            followup_suggestion=exit_output.followup_suggestion,
            error_from_pipeline=(
                exit_output.error_from_sql_worker
                or exit_output.error_from_analysis_worker
                or exit_output.error_from_plot_worker
                or exit_output.error_from_followup_worker
            ),
        )

    def __call__(
        self,
        question: str,
        input_dataframe: Optional[dict] = None,
        server_mode: Optional[bool] = False,
    ) -> Union[ExitWorkerOutput, AgentOutput]:
        if server_mode:
            kwargs = self.route_worker_kwargs.get("plot", None)
            kwargs = (
                {"plot_image": False}
                if kwargs is None
                else {**kwargs, "plot_image": False}
            )
            self.route_worker_kwargs["plot"] = kwargs

        output = self.run(question=question, input_dataframe=input_dataframe)
        # TODO: Watch out dict here type mismatch with run
        self.history.push(output=output)
        if server_mode:
            output = self.convert_exit_output_to_agent_output(exit_output=output)
        return output



import re 
import os 
import pandas as pd 
import kagglehub
import sqlite3
from pathlib import Path
from platformdirs import user_cache_dir
from premsql.logger import setup_console_logger

logger = setup_console_logger("[FRONTEND-UTILS]")

def _is_valid_kaggle_id(kaggle_id: str) -> bool:
    pattern = r'^[a-zA-Z0-9_-]+/[a-zA-Z0-9_-]+$'
    return bool(re.match(pattern, kaggle_id))

def download_from_kaggle(kaggle_dataset_id: str):
    path = kagglehub.dataset_download(handle=kaggle_dataset_id)
    return path 

def _migrate_to_sqlite(csv_folder: Path, sqlite_db_path: Path) -> Path:
    """Common migration logic for both Kaggle and local CSV uploads."""
    conn = sqlite3.connect(sqlite_db_path)
    try:
        for csv_file in csv_folder.glob('*.csv'):
            table_name = csv_file.stem
            df = pd.read_csv(csv_file)
            df.to_sql(table_name, conn, if_exists='replace', index=False)
            logger.info(f"Migrated {csv_file.name} to table '{table_name}'")
        
        logger.info(f"Successfully migrated all CSV files to {sqlite_db_path}")
        return sqlite_db_path
    except Exception as e:
        logger.error(f"Error during migration: {e}")
        raise
    finally:
        conn.close()

def migrate_from_csv_to_sqlite(
    folder_containing_csvs: str, 
    session_name: str
) -> Path:
    sqlite_db_folder = Path(user_cache_dir()) / "premsql" / "kaggle"
    os.makedirs(sqlite_db_folder, exist_ok=True)
    sqlite_db_path = sqlite_db_folder / f"{session_name}.sqlite"
    return _migrate_to_sqlite(Path(folder_containing_csvs), sqlite_db_path)

def migrate_local_csvs_to_sqlite(
    uploaded_files: list,
    session_name: str
) -> Path:
    cache_dir = Path(user_cache_dir())
    csv_folder = cache_dir / "premsql" / "csv_uploads" / session_name
    sqlite_db_folder = cache_dir / "premsql" / "csv_uploads"
    
    os.makedirs(csv_folder, exist_ok=True)
    os.makedirs(sqlite_db_folder, exist_ok=True)
    
    sqlite_db_path = sqlite_db_folder / f"{session_name}.sqlite"
    
    # Save uploaded files to CSV folder
    for uploaded_file in uploaded_files:
        file_path = csv_folder / uploaded_file.name
        with open(file_path, 'wb') as f:
            f.write(uploaded_file.getvalue())
    
    return _migrate_to_sqlite(csv_folder, sqlite_db_path)


import json
from pathlib import Path
from typing import Optional, Sequence

from tqdm.auto import tqdm

from premsql.datasets.base import (
    SupervisedDatasetForTraining,
    Text2SQLBaseDataset,
    Text2SQLBaseInstance,
)
from premsql.evaluator.base import BaseExecutor, Text2SQLEvaluator
from premsql.generators.base import Text2SQLGeneratorBase
from premsql.logger import setup_console_logger
from premsql.prompts import ERROR_HANDLING_PROMPT

logger = setup_console_logger("[ERROR-HANDLING-DATASET]")


class ErrorDatasetInstance(Text2SQLBaseInstance):

    def __init__(self, dataset: list[dict]) -> None:
        super().__init__(dataset=dataset)

    def apply_prompt(self, prompt_template: Optional[str] = ERROR_HANDLING_PROMPT):
        data_to_return = []
        for content in tqdm(
            self.dataset, total=len(self.dataset), desc="Applying error prompt"
        ):
            assert "error" in content, "key error is not present"
            error_msg = content["error"]

            if error_msg is not None:
                prompt = content["prompt"].split("# SQL:")[0].strip()
                prediction = content["generated"]
                error_prompt = prompt_template.format(
                    existing_prompt=prompt, sql=prediction, error_msg=error_msg
                )
                data_to_return.append(
                    {
                        "db_id": content["db_id"],
                        "question": content["question"],
                        "SQL": content["SQL"],
                        "prompt": error_prompt,
                        "db_path": content["db_path"],
                    }
                )
        return data_to_return


class ErrorDatasetGenerator:
    @classmethod
    def from_existing(
        cls,
        experiment_name: str,
        experiment_folder: Optional[str] = None,
        tokenize_model_name_or_path: Optional[str] = None,
        hf_token: Optional[str] = None,
    ) -> dict:
        experiment_folder = Path("./experiments") or Path(experiment_folder)
        experiment_path = (
            experiment_folder / "train" / experiment_name / "error_dataset.json"
        )
        if not experiment_path.exists():
            raise FileNotFoundError(f"Path {experiment_path} does not exists")
        dataset = json.load(open(experiment_path, "r"))
        return (
            ErrorDatasetInstance(dataset=dataset)
            if not tokenize_model_name_or_path
            else SupervisedDatasetForTraining(
                dataset=dataset,
                model_name_or_path=tokenize_model_name_or_path,
                hf_token=hf_token,
            )
        )

    def __init__(
        self,
        generator: Text2SQLGeneratorBase,
        executor: BaseExecutor,
    ):
        self.generator = generator
        self.evaluator = Text2SQLEvaluator(
            executor=executor, experiment_path=self.generator.experiment_path
        )

    def generate_and_save(
        self,
        datasets: Sequence[Text2SQLBaseDataset],
        path_to_save: Optional[str] = None,
        force: Optional[bool] = False,
        tokenize: Optional[bool] = False,
        prompt_template: Optional[str] = ERROR_HANDLING_PROMPT,
        hf_token: Optional[str] = None,
    ) -> None:

        path_to_save = (
            (self.generator.experiment_path / "error_dataset.json")
            if path_to_save is None
            else Path(path_to_save)
        )
        if path_to_save.exists() and force == False:
            logger.info("Error dataset already exists")
            with open(path_to_save, "r") as json_file:
                data_to_return = json.load(json_file)
            return data_to_return

        responses = self.generator.generate_and_save_results(
            dataset=datasets, temperature=0.1, max_new_tokens=256, force=force
        )
        logger.info("Starting Evaluation")
        _ = self.evaluator.execute(
            metric_name="accuracy",
            model_responses=responses,
        )
        del responses

        # Now iterate over the error dataset
        with open(self.generator.experiment_path / "predict.json", "r") as file:
            error_dataset = json.load(file)

        error_instances = ErrorDatasetInstance(dataset=error_dataset).apply_prompt(
            prompt_template=prompt_template
        )

        with open(path_to_save, "w") as json_file:
            json.dump(error_instances, json_file, indent=4)

        return (
            error_instances
            if not tokenize
            else SupervisedDatasetForTraining(
                dataset=error_instances,
                model_name_or_path=self.generator.model_name_or_path,
                hf_token=hf_token,
            )
        )